#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>

#if ENABLE_NLS
#include <libintl.h>
#define _(String) gettext( String )
#else
#define _(String) String
#endif
#define N_(String) String
#if HAVE_SETLOCALE
#include <locale.h>
#endif

#include <getopt.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <dirent.h>
#if HAVE_SYS_DIRENT_H
#include <sys/dirent.h>
#endif
#if HAVE_PATHS_H
#include <paths.h>
#include <sys/fcntl.h>
#else
#define _PATH_RWHODIR "/var/spool/rwho"
#include <sys/file.h>
#endif
#include <protocols/rwhod.h>

#define USERINC	16

typedef struct _tUser {
	char	hostname[32];
	struct	whoent	we;
} tUser;

enum 	sort_method { sHost, sUser, sTime, sIdle, s_Last };
char	*sortoption[] = { "hostname", "username", "ontime", "idletime" };

static	tUser			*Users = NULL;
static	int			nrusers = 0;
static	int			maxusers = 0;
static	int			show_help;
static	int			show_version;
static	enum	sort_method	sort_method = sUser;
static	int			sort_reverse;
static	int			show_all;
static	char			*progname;
static	int			width;

static void readusers(void);
static void sortusers(void);
static void showusers(void);

static
void usage(int status)
{
	if (status != 0) {
		fprintf(stderr,_("Try `%s --help' for more information.\n"),
			progname);
	}
	else {
		printf(_("Usage: %s [OPTION]...\n"), progname);
		printf(_("Show who is logged in on remote machines.\n"
			"\n"
			"  -a  --show-all       show all users, instead of the not-idle ones\n"
			"  -h  --sort=hostname  sort according to the hostname\n"
			"  -u  --sort=username  sort according to the number of users\n"
			"  -t  --sort=ontime    sort according to the ontime\n"
			"  -i  --sort=idletime  sort according to the idletime\n"
			"  -r  --sort-reverse   reverse the sort order\n"
			"      --help           display this help and exit\n"
			"      --version        output version information and exit\n"
			"\n"
			"Users are assumed idle when they have been inactive for more than\n"
			"an hour. Remote hosts for which no data is received in the last 11\n"
			"minutes are considered down. No idle time is displayed when the user\n"
			"is at least active for a minute.\n"));
	}
	exit(status);
}

int main(int argc, char *argv[])
{
static struct option longopt[] = {
	{ "help", no_argument, &show_help, 1},
	{ "version", no_argument, &show_version, 1},
	{ "all", no_argument, NULL, 'a'},
	{ "sort", required_argument, NULL, 1},
	{ "sort-reverse", no_argument, NULL, 'r'},
	{ NULL } };
	int	ch;

#if HAVE_SETLOCALE
	setlocale(LC_MESSAGES, "");
#endif
#if ENABLE_NLS
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);
#endif

	progname = strrchr(argv[0],'/');
	if (progname) progname++;
	else progname = argv[0];

	while ((ch=getopt_long(argc,argv,"ahitur",longopt,NULL)) != -1)
	 switch (ch) {
	  case   0 :	break;		/* long option */
	  		/* --sort= */
	  case   1 :	for (ch=0; ch<s_Last; ch++)
				if (strcmp(optarg,sortoption[ch])==0) {
					sort_method = (enum sort_method)ch;
					break;
				}
			if (ch==s_Last) {
				fprintf(stderr,_("%s: invalid sort option -- %s\n"),
					progname, optarg);
				usage(1);
			}
			break;
	  case 'a' :	show_all++;
			break;
	  case 'r' :	sort_reverse++;
			break;
	  case 'h' :	sort_method = sHost;
			break;
	  case 'u' :	sort_method = sUser;
			break;
	  case 'i' :	sort_method = sIdle;
			break;
	  case 't' :	sort_method = sTime;
			break;
	  default  :	usage(1);
	 }

	if (show_version) {
		printf("%s - %s\n", progname, PACKAGE_VERSION);
		exit(0);
	}

	if (show_help)
		usage(0);

	readusers();
	sortusers();
	showusers();
	return 0;
}

static
void readusers()
{
	DIR	*dir;
	struct	dirent	*dirp;
	time_t	uptime;
#if HAVE_GETDENTS
	int	len;
	struct	dirent dirs[32];
#endif

	if (chdir(_PATH_RWHODIR) != 0) {
		perror(_PATH_RWHODIR);
		exit(1);
	}
	dir = opendir(".");
	if (dir==NULL) {
		perror("opendir");
		exit(1);
	}
	width = 8;
	time(&uptime);
	uptime -= 11*60;
#if HAVE_GETDENTS
while ((len=getdents(dirfd(dir),dirs,sizeof(dirs)))>0)
	for (dirp=dirs; dirp!=((char *)dirs)+len; ((char *)dirp) += dirp->d_reclen) {
#else
	while ((dirp=readdir(dir)) != NULL) {
#endif
		struct	whod	rwho;
		int	fd;
		int	size;
		time_t	timediff;

		if (strcmp(dirp->d_name,".")==0 ||
		    strcmp(dirp->d_name,"..")==0)
			continue;

		fd = open(dirp->d_name,O_RDONLY);
		if (fd<0) {
			perror(dirp->d_name);
			continue;
		}
		size = read(fd,&rwho,sizeof(struct whod));
		close(fd);

		/* did we read enough basic data? */
		if (size<60)
			continue;

		/* host down? */
		if (rwho.wd_recvtime < uptime)
			continue;

		/* add the users to the user list */
		timediff = rwho.wd_sendtime - rwho.wd_recvtime;
		for (fd=0; size>60; fd++, size-= sizeof(struct whoent))
			if (show_all || rwho.wd_we[fd].we_idle<3600) {
				int w = strlen(rwho.wd_hostname)+strlen(rwho.wd_we[fd].we_utmp.out_line)+1;
				if (nrusers == maxusers) {
					maxusers += USERINC;
					Users = realloc(Users,maxusers*sizeof(tUser));
					if (Users == NULL) {
						fprintf(stderr, _("%s: too many users\n"),progname);
						exit(2);
					}
				}
				if (w>width) width = w;
				Users[nrusers].we = rwho.wd_we[fd];
				Users[nrusers].we.we_utmp.out_time += timediff;
				strcpy(Users[nrusers].hostname, rwho.wd_hostname);
				nrusers++;
			}
	}
	closedir(dir);
}

static
int sortuser(const void *aa, const void *bb)
{
	const	tUser *a = (tUser *)aa;
	const	tUser *b = (tUser *)bb;
	int	res = 0; /* stupid GCC: it is _always_ defined in the switch!! */

	switch (sort_method) {
	 case sUser :
	 	/* use strncmp instead of memcmp in case of strange locale */
		res = strncmp(a->we.we_utmp.out_name, b->we.we_utmp.out_name, sizeof(a->we.we_utmp.out_name));
		break;
	 case sHost :
		res = strcmp(a->hostname, b->hostname);
		break;
	 case sIdle :
		res = b->we.we_idle - a->we.we_idle;
		break;
	 case sTime :
		res = b->we.we_utmp.out_time - a->we.we_utmp.out_time;
		break;
	 case s_Last :
		res = 0;
		break;
	}

	return (sort_reverse) ? -res : res;
}

static
void sortusers()
{
	qsort(Users,nrusers,sizeof(tUser),sortuser);
}

static
void showusers()
{
static	char	*szMon[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
			    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
	int	i;

	for (i=0; i<nrusers; i++) {
		char	name[32+8+1];
		char	line[80+1];
		tUser	*pu = &Users[i];
		struct	tm	*tm;
		char	*p = &line[0];

		sprintf(name,"%s:%s",pu->hostname,pu->we.we_utmp.out_line);
		tm = localtime(&pu->we.we_utmp.out_time);

		sprintf(p,"%-8.8s %*s %s %2d %02d:%02d ",
			pu->we.we_utmp.out_name,
			-width, name,
			szMon[tm->tm_mon],
			tm->tm_mday,
			tm->tm_hour,
			tm->tm_min);
		p += strlen(p);
		if (pu->we.we_idle>99*3600+59*60+59)
			pu->we.we_idle = 99*3600+59*60+59;
		if (pu->we.we_idle>=3600) {
			sprintf(p,"%2d",pu->we.we_idle/3600);
			p += strlen(p);
		}
		if (pu->we.we_idle>=60) {
			sprintf(p,":%02d",pu->we.we_idle/60%60);
			p += strlen(p);
		}
		puts(line);
	}
}
